""".. include:: ../../README.md"""  # noqa: D415

from cmi_docx.document import ExtendDocument  # noqa: F401
from cmi_docx.paragraph import ExtendParagraph, FindParagraph  # noqa: F401
from cmi_docx.run import ExtendRun, FindRun  # noqa: F401
from cmi_docx.styles import ParagraphStyle, RunStyle, TableStyle  # noqa: F401
from cmi_docx.table import ExtendCell  # noqa: F401



"""Module for extending python-docx Paragraph objects."""

import bisect
import dataclasses
import itertools
import re

from docx.text import paragraph as docx_paragraph
from docx.text import run as docx_run

from cmi_docx import run, styles


@dataclasses.dataclass
class FindParagraph:
    """Data class for maintaing find results in paragraphs.

    Attributes:
        paragraph: The paragraph containing the text.
        character_indices: A list of matching character indices of the text in the
            paragraph.
    """

    paragraph: docx_paragraph.Paragraph
    character_indices: list[tuple[int, int]]


class ExtendParagraph:
    """Extends a python-docx Word paragraph with additional functionality."""

    def __init__(self, paragraph: docx_paragraph.Paragraph) -> None:
        """Initializes an ExtendParagraph object.

        Args:
            paragraph: The paragraph to extend.
        """
        self.paragraph = paragraph

    def find_in_paragraph(self, needle: str) -> FindParagraph:
        """Finds the indices of a text relative to the paragraph.

        Args:
            needle: The text to find.

        Returns:
            The indices of the text in the paragraph.
        """
        within_paragraph_indices = [
            (match.start(), match.end())
            for match in re.finditer(re.escape(needle), self.paragraph.text)
        ]

        return FindParagraph(
            paragraph=self.paragraph,
            character_indices=within_paragraph_indices,
        )

    def find_in_runs(self, needle: str) -> list[run.FindRun]:
        """Finds the indices of a text relative to the paragraph's runs.

        Args:
            needle: The text to find.

        Returns:
            The indices of the text in the paragraph.
        """
        if len(needle) == 0:
            return []

        run_finds: list[run.FindRun] = []
        run_lengths = [len(run.text) for run in self.paragraph.runs]
        cumulative_run_lengths = list(itertools.accumulate(run_lengths))
        for occurence in self.find_in_paragraph(needle).character_indices:
            start_run = bisect.bisect_right(cumulative_run_lengths, occurence[0])
            end_run = bisect.bisect_right(
                cumulative_run_lengths[:-1], occurence[1], lo=start_run
            )

            start_index = (
                occurence[0] - cumulative_run_lengths[start_run - 1]
                if start_run > 0
                else occurence[0]
            )
            end_index = (
                occurence[1] - cumulative_run_lengths[end_run - 1]
                if end_run > 0
                else occurence[1]
            )

            run_finds.append(
                run.FindRun(
                    paragraph=self.paragraph,
                    run_indices=(start_run, end_run),
                    character_indices=(start_index, end_index),
                )
            )
        return run_finds

    def replace(
        self, needle: str, replace: str, style: styles.RunStyle | None = None
    ) -> None:
        """Finds and replaces text in a Word paragraph.

        Args:
            needle: The text to find.
            replace: The text to replace.
            style: The style to apply to the replacement text.
        """
        run_finder = self.find_in_runs(needle)
        run_finder.sort(
            key=lambda x: (x.run_indices[0], x.character_indices[0]), reverse=True
        )

        for run_find in run_finder:
            run_find.replace(replace, style)

    def insert_run(self, index: int, text: str, style: styles.RunStyle) -> docx_run.Run:
        """Inserts a run into a paragraph.

        Args:
            index: The index to insert the run at.
            text: The text of the run.
            style: The style of the run, see run.ExtendRun.format for more details.

        Returns:
            The inserted run.
        """
        if index == len(self.paragraph.runs):
            self.paragraph.add_run(text)
        else:
            new_run = self.paragraph._element._new_r()
            new_run.text = text
            if index < 0:
                self.paragraph.runs[index]._element.addnext(new_run)
            else:
                self.paragraph.runs[index]._element.addprevious(new_run)

        run.ExtendRun(self.paragraph.runs[index]).format(style)
        return self.paragraph.runs[index]

    def format(
        self,
        style: styles.ParagraphStyle,
    ) -> None:
        """Formats a paragraph in a Word document.

        Args:
            style: The style to apply to the paragraph.
        """
        if style.line_spacing is not None:
            self.paragraph.paragraph_format.line_spacing = style.line_spacing

        if style.alignment is not None:
            self.paragraph.alignment = style.alignment

        if style.space_before is not None:
            self.paragraph.paragraph_format.space_before = style.space_before

        if style.space_after is not None:
            self.paragraph.paragraph_format.space_after = style.space_after

        for paragraph_run in self.paragraph.runs:
            run.ExtendRun(paragraph_run).format(
                styles.RunStyle(
                    bold=style.bold,
                    italic=style.italic,
                    font_size=style.font_size,
                    font_rgb=style.font_rgb,
                )
            )



"""Style interfaces for document properties."""

import dataclasses

from docx.enum import text


@dataclasses.dataclass
class RunStyle:
    """Dataclass for run style arguments."""

    bold: bool | None = None
    italic: bool | None = None
    underline: bool | None = None
    superscript: bool | None = None
    subscript: bool | None = None
    font_size: int | None = None
    font_rgb: tuple[int, int, int] | None = None


@dataclasses.dataclass
class ParagraphStyle:
    """Dataclass for paragraph style arguments."""

    bold: bool | None = None
    italic: bool | None = None
    font_size: int | None = None
    font_rgb: tuple[int, int, int] | None = None
    line_spacing: float | None = None
    space_before: float | None = None
    space_after: float | None = None
    alignment: text.WD_PARAGRAPH_ALIGNMENT | None = None


@dataclasses.dataclass
class TableStyle:
    """Dataclass for table style arguments."""

    paragraph: ParagraphStyle | None = None
    space_before: float | None = None
    space_after: float | None = None
    background_rgb: tuple[int, int, int] | None = None



"""Extends a python-docx Table cell with additional functionality."""

from docx import oxml, table
from docx.oxml import ns

from cmi_docx import paragraph, styles


class ExtendCell:
    """Extends a python-docx Word cell with additional functionality."""

    def __init__(self, cell: table._Cell) -> None:
        """Initializes an ExtendCell object.

        Args:
            cell: The cell to extend.
        """
        self.cell = cell

    def format(self, style: styles.TableStyle) -> None:
        """Formats a cell in a Word table.

        Args:
            style: The style to apply to the cell.
        """
        if style.paragraph is not None:
            for table_paragraph in self.cell.paragraphs:
                paragraph.ExtendParagraph(table_paragraph).format(style.paragraph)

        if style.background_rgb is not None:
            shading = oxml.parse_xml(
                (
                    r'<w:shd {} w:fill="'
                    + f"{rgb_to_hex(*style.background_rgb)}"
                    + r'"/>'
                ).format(
                    ns.nsdecls("w"),
                ),
            )
            self.cell._tc.get_or_add_tcPr().append(shading)  # noqa: SLF001


def rgb_to_hex(r: int, g: int, b: int) -> str:
    """Converts RGB values to a hexadecimal color code.

    Args:
        r: The red component of the RGB color.
        g: The green component of the RGB color.
        b: The blue component of the RGB color.

    Returns:
        The hexadecimal color code representing the RGB color.
    """
    return f"#{r:02x}{g:02x}{b:02x}".upper()



"""Module for extending python-docx Run objects."""

from docx import shared
from docx.enum import text
from docx.text import paragraph as docx_paragraph

from cmi_docx import styles


class FindRun:
    """Data class for maintaing find results in runs.

    Attributes:
        paragraph: The paragraph containing the text.
        run_indices: The run indices of the text needle's start and end.
        character_indices: The character indices of the text in the runs.
            The first index is the start of the text in the first run containing
            the text. The second index is the end of the text in the last run.
    """

    def __init__(
        self,
        paragraph: docx_paragraph.Paragraph,
        run_indices: tuple[int, int],
        character_indices: tuple[int, int],
    ) -> None:
        """Initializes a FindRun object.

        Args:
            paragraph: The paragraph containing the text.
            run_indices: The run indices of the text needle's start and end.
            character_indices: The character indices of the text in the runs.
        """
        self.paragraph = paragraph
        self.run_indices = run_indices
        self.character_indices = character_indices
        self._replacement_done = False

    @property
    def runs(self) -> list[docx_paragraph.Run]:
        """Returns the runs containing the text."""
        return self.paragraph.runs[self.run_indices[0] : self.run_indices[1] + 1]

    def replace(self, replace: str, style: styles.RunStyle | None = None) -> None:
        """Replaces the text in the runs with the replacement text.

        Args:
            replace: The text to replace.
            style: The style to apply to the replacement text. If None, the
                replacement text will have the same style as the first character of the
                original text.
        """
        if self._replacement_done:
            msg = "Cannot use a FindRun replacement more than once."
            raise ValueError(msg)

        if style is None:
            self._replace_without_style(replace)
        else:
            self._replace_with_style(replace, style)
        self._replacement_done = True
        return

    def _replace_without_style(self, replace: str) -> None:
        """Replaces the text in the runs with the replacement text.

        Args:
            replace: The text to replace.
        """
        start = self.character_indices[0]
        end = self.character_indices[1]

        if len(self.runs) == 1:
            self.runs[0].text = (
                self.runs[0].text[:start] + replace + self.runs[0].text[end:]
            )
        else:
            self.runs[0].text = self.runs[0].text[:start] + replace
            for run in self.runs[1:-1]:
                run.clear()
            self.runs[-1].text = self.runs[-1].text[end:]

    def _replace_with_style(self, replace: str, style: styles.RunStyle) -> None:
        """Replaces the text in the runs with the replacement text and style.

        Args:
            replace: The text to replace.
            style: The style to apply to the replacement text.
        """
        start = self.character_indices[0]
        end = self.character_indices[1]

        pre, post = self.runs[0].text[:start], self.runs[-1].text[end:]
        self.runs[0].text = pre
        for index in range(1, len(self.runs)):
            self.runs[index].text = ""

        new_run = self.paragraph._element._new_r()
        new_run.text = replace
        self.paragraph.runs[self.run_indices[0]]._element.addnext(new_run)
        ExtendRun(self.paragraph.runs[self.run_indices[0] + 1]).format(style)

        post_run = self.paragraph._element._new_r()
        post_run.text = post
        self.paragraph.runs[self.run_indices[0] + 1]._element.addnext(post_run)
        pre_style = ExtendRun(self.paragraph.runs[self.run_indices[0]]).get_format()
        ExtendRun(self.paragraph.runs[self.run_indices[0] + 2]).format(pre_style)

    def __lt__(self, other: "FindRun") -> bool:
        """Sorts FindRun in order of appearance in the paragraph.

        Makes FindRun objects sortable.

        Args:
            other: The other FindRun object.

        Returns:
            True if the character index of the first run is less than the
            character index of the other run.
        """
        if self.paragraph != other.paragraph:
            msg = "Cannot compare FindRun objects from different paragraphs."
            raise ValueError(msg)

        if self.run_indices[0] == other.run_indices[0]:
            return self.character_indices[0] < other.character_indices[0]
        return self.run_indices[0] < other.run_indices[0]


class ExtendRun:
    """Extends a python-docx Word run with additional functionality."""

    def __init__(self, run: docx_paragraph.Run) -> None:
        """Initializes an ExtendRun object.

        Args:
            run: The run to extend.
        """
        self.run = run

    def format(self, style: styles.RunStyle) -> None:
        """Formats a run in a Word document.

        Args:
            style: The style to apply to the run.
        """
        if style.superscript and style.subscript:
            msg = "Cannot have superscript and subscript at the same time."
            raise ValueError(msg)

        if style.bold is not None:
            self.run.bold = style.bold
        if style.italic is not None:
            self.run.italic = style.italic
        if style.underline is not None:
            self.run.underline = style.underline
        if style.superscript is not None:
            self.run.font.superscript = style.superscript
        if style.subscript is not None:
            self.run.font.subscript = style.subscript
        if style.font_size is not None:
            self.run.font.size = shared.Pt(style.font_size)
        if style.font_rgb is not None:
            self.run.font.color.rgb = shared.RGBColor(*style.font_rgb)

    def get_format(self) -> styles.RunStyle:
        """Returns the formatting of the run.

        Returns:
            The formatting of the run.
        """
        underline: bool | None
        if isinstance(self.run.underline, text.WD_UNDERLINE):
            underline = self.run.underline != text.WD_UNDERLINE.NONE
        else:
            underline = self.run.underline

        return styles.RunStyle(
            bold=self.run.bold,
            italic=self.run.italic,
            underline=underline,
            superscript=self.run.font.superscript,
            subscript=self.run.font.subscript,
            font_size=self.run.font.size,
            font_rgb=self.run.font.color.rgb,
        )
